/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 *
 * Authors:  Arif Husen <arifhrashid@gmail.com>
 *           COMSATS University Islamabad, Lahore Campus
 */
#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <algorithm>
#include <sstream>
#include <ostream>
#include <map>
#include <iterator>
#include "ns3/log.h"
#include "ns3/object-factory.h"
#include "ns3/queue.h"
#include "ns3/ipv4-queue-disc-item.h"
#include "ns3/socket.h"
#include "ns3/packet.h"
#include "tips-queue-disc.h"
#include "ns3/ppp-header.h"
#include "ns3/ipv4-header.h"
#include "ns3/simulator.h"
using namespace ns3;
namespace ns3
{
NS_LOG_COMPONENT_DEFINE ("TipsQueueDisc");
NS_OBJECT_ENSURE_REGISTERED (TipsQueueDisc);
TypeId TipsQueueDisc::GetTypeId (void)
{
    static TypeId tid = TypeId ("ns3::TipsQueueDisc")
                        .SetParent<QueueDisc> ()
                        .SetGroupName ("TrafficControl")
                        .AddConstructor<TipsQueueDisc> ()
                        .AddAttribute ("MaxSize",
                                       "The maximum number of packets accepted by this queue disc.",
                                       QueueSizeValue (QueueSize ("1000p")),
                                       MakeQueueSizeAccessor (&QueueDisc::SetMaxSize,
                                               &QueueDisc::GetMaxSize),
                                       MakeQueueSizeChecker ())
                        ;
    return tid;
}
TipsQueueDisc::TipsQueueDisc () : QueueDisc (QueueDiscSizePolicy::MULTIPLE_QUEUES, QueueSizeUnit::PACKETS)
{
    NS_LOG_FUNCTION (this);
}
TipsQueueDisc::~TipsQueueDisc ()
{
    NS_LOG_FUNCTION (this);
}
int TipsQueueDisc::load_learner_output()
{
    NS_LOG_LOGIC ("Loading Machine Learning......");
    std::ifstream erlang_profiles_src;
    std::string ng1 = "g1";
    std::string ng2 = "g2";
    std::string ng3 = "g3";
    std::string nc = "c";
    if (this->tips_NodeID.compare(ng1) == 0)
    {
        erlang_profiles_src.open("g1_ml.txt");
    }
    else if (this->tips_NodeID.compare(ng2) == 0)
    {
        erlang_profiles_src.open("g2_ml.txt");
    }
    else if (this->tips_NodeID.compare(ng3) == 0)
    {
        erlang_profiles_src.open("g3_ml.txt");
    }
    else if (this->tips_NodeID.compare(nc) == 0)
    {
        erlang_profiles_src.open("c_ml.txt");
    }
    if (erlang_profiles_src.is_open())
    {
        std::string tp;
        int tp_no = 0;
        while(getline(erlang_profiles_src, tp))
        {
            if (tp_no > 0)
            {
                std::stringstream line_str(tp);
                while(line_str.good())
                {
                    std::string time_str;
                    std::string node_str;
                    std::string queues_str;
                    std::string erlang_str;
                    std::string weights_str;
                    getline(line_str, time_str, ',');
                    getline(line_str, node_str, ',');
                    getline(line_str, queues_str, ',');
                    getline(line_str, erlang_str, ',');
                    std::vector<int> erl_ints = this->take_int(erlang_str);
                    for(int i = 0; i < (int)erl_ints.size(); i++)
                    {
                        int erl_index = i + 1;
                        this->m_erlang_profiles[time_str][node_str][queues_str][erl_index] = erl_ints[i];
                    }
                    getline(line_str, weights_str, ',');
                    std::vector<int> weights_ints = take_int(weights_str);
                    std::sort(weights_ints.begin(), weights_ints.end(), std::greater<int>());
                    for(int i = 0; i < (int)weights_ints.size(); i++)
                    {
                        int weight_index = i + 1;
                        int w = weights_ints[i];
                        this->m_erlang_weights[time_str][node_str][queues_str][weight_index] = w;
                    }
                }
            }
            tp_no++;
        }
    }
    erlang_profiles_src.close();
    return 0;
}
std::vector<int> TipsQueueDisc::take_int(std::string str)
{
    std::stringstream ss(str);
    std::vector<int> result;
    char ch;
    int tmp;
    while(ss >> tmp)
    {
        result.push_back(tmp);
        ss >> ch;
    }
    return result;
}
bool TipsQueueDisc::detect_change(std::string t)
{
    bool b = false;
    if (m_erlang_profiles.find(t) == m_erlang_profiles.end())
    {
        b = false;
    }
    else
    {
        b = true;
    }
    return b;
}
int TipsQueueDisc::get_erlang(std::string t,std::string n,std::string s,int i)
{
    int erl_int;
    erl_int = m_erlang_profiles[t][n][s][i];
    return erl_int;
}
int TipsQueueDisc::get_weight(std::string t,std::string n,std::string s,int i)
{
    int erl_int;
    erl_int = m_erlang_weights[t][n][s][i];
    return erl_int;
}
int TipsQueueDisc::get_sources(std::string t,std::string n)
{
    int s_int = 3;
    if (m_erlang_profiles.find(t) == m_erlang_profiles.end())
    {
    }
    else
    {
        auto src_itr = m_erlang_profiles[t][n].begin();
        std::string s =  src_itr->first;
        const char *cstr = s.c_str();
        sscanf(cstr,"%d",&s_int);
    }
    return s_int;
}
std::string TipsQueueDisc::get_sources_str(std::string t,std::string n)
{
    std::string s = "4";
    auto it = m_erlang_profiles[t][n].begin();
    if (it != m_erlang_profiles[t][n].end())
    {
        s = it->first;
    }
    return s;
}
bool TipsQueueDisc::update_erlOrder(std::string tsss)
{
    int s_int = 3;
    std::string s = "3";
    std::string n = tips_NodeID;
    auto src_itr = m_erlang_profiles[tsss][n].begin();
    if (src_itr != m_erlang_profiles[tsss][n].end())
    {
        s =  src_itr->first;
        const char *cstr = s.c_str();
        sscanf(cstr,"%d",&s_int);
    }
    std::map<int,int> erl = m_erlang_profiles[tsss][n][s];
    std::map<int, int>::iterator it;
    int e_index = 0;
    int e_val = 0;
    for (it = erl.begin(); it != erl.end(); it++)
    {
        e_index = it->first - 1;
        e_val = it->second;
        prio2band[e_index] = e_val;
    }
    prio2band[e_index+1] = 4;
    return true;
}
bool TipsQueueDisc::DoEnqueue (Ptr<QueueDiscItem> item)
{
    NS_LOG_FUNCTION (this << item);
    int ts = (int) Simulator::Now().GetSeconds();
    std::ostringstream toss;
    toss << ts;
    std::string time_str = toss.str();
    this->tips_NodeID = this->qd_nodeID;
    if (this->detect_change(time_str) == true)
    {
        this->update_erlOrder(time_str);
    }
    if (GetCurrentSize () >= GetMaxSize ())
    {
        NS_LOG_LOGIC ("Queue disc limit exceeded -- dropping packet");
        DropBeforeEnqueue (item, LIMIT_EXCEEDED_DROP);
        return false;
    }
    Ptr<const Ipv4QueueDiscItem> ipItem = DynamicCast<const Ipv4QueueDiscItem>(item);
    const Ipv4Header ipHeader = ipItem->GetHeader();
    Ipv4Address i4ad = ipHeader.GetSource();
    std::stringstream holdBuff;
    holdBuff << i4ad;
    std::string str_a;
    getline(holdBuff, str_a, '.');
    std::string str_b;
    getline(holdBuff, str_b, '.');
    std::string str_c;
    getline(holdBuff, str_c, '.');
    std::string str_d;
    getline(holdBuff, str_d, '.');
    int a,b,c,d;
    std::istringstream ( str_a ) >> a;
    std::istringstream ( str_b ) >> b;
    std::istringstream ( str_c ) >> c;
    std::istringstream ( str_d ) >> d;
    NS_LOG_LOGIC ("ip " << a<<b << c<< d);
    int p = (c%4)-1;
    if (p < 0){
    p = 3;
    }
    int band = 0; 
if (p < 13) {
  band = prio2band[p];
}
    bool retval = GetInternalQueue (band)->Enqueue (item);
    if (!retval)
    {
        NS_LOG_WARN ("Packet enqueue failed. Check the size of the internal queues");
    }
    NS_LOG_LOGIC ("Number packets band " << band << ": " << GetInternalQueue (band)->GetNPackets ());
    return retval;
}
Ptr<QueueDiscItem> TipsQueueDisc::DoDequeue (void)
{
    NS_LOG_FUNCTION (this);
    Ptr<QueueDiscItem> item;
    for (uint32_t i = 0; i < GetNInternalQueues (); i++)
    {
        int ts = (int)Simulator::Now().GetSeconds();
        std::ostringstream toss;
        toss << ts;
        std::string time_str = toss.str();
        std::string s = "4";
        std::string n = this->tips_NodeID;
        auto it = m_erlang_profiles[time_str][n].begin();
        if (it != m_erlang_profiles[time_str][n].end())
        {
            s = it->first;
        }
        int w = 0;
        if (i == 4) {
        w = 4;
        } else {
        w = this->get_weight(time_str,n,s,i);
        }
        
        for (int n=0; n<w; n++)
        {
            if ((item = GetInternalQueue (i)->Dequeue ()) != 0)
            {
                NS_LOG_LOGIC ("Popped from Q " << i << ": " << item);
                NS_LOG_LOGIC ("Number packets Q " << i << ": " << GetInternalQueue (i)->GetNPackets ());
                return item;
            }
        }
    }
    NS_LOG_LOGIC ("Queue empty");
    return item;
}
Ptr<const QueueDiscItem> TipsQueueDisc::DoPeek (void)
{
    NS_LOG_FUNCTION (this);
    Ptr<const QueueDiscItem> item;
    for (uint32_t i = 0; i < GetNInternalQueues (); i++)
    {
        if ((item = GetInternalQueue (i)->Peek ()) != 0)
        {
            NS_LOG_LOGIC ("Peeked from Q " << i << ": " << item);
            NS_LOG_LOGIC ("Number packets Q " << i << ": " << GetInternalQueue (i)->GetNPackets ());
            return item;
        }
    }
    NS_LOG_LOGIC ("Queue empty");
    return item;
}
bool TipsQueueDisc::CheckConfig (void)
{
    NS_LOG_FUNCTION (this);
    int ts = (int) Simulator::Now().GetSeconds();
        std::ostringstream tss;
        tss << ts;
        std::string tsss = tss.str();
        this->tips_NodeID = this->qd_nodeID;
        size_t Q = 4;
        load_learner_output();
    if (GetNQueueDiscClasses () > 0)
    {
        NS_LOG_ERROR ("TipsQueueDisc cannot have classes");
        return false;
    }
    if (GetNPacketFilters () != 0)
    {
        NS_LOG_ERROR ("TipsQueueDisc needs no packet filter");
        return false;
    }
    if (GetNInternalQueues () == 0)
    {
        ObjectFactory factory;
        factory.SetTypeId ("ns3::DropTailQueue<QueueDiscItem>");
        factory.Set ("MaxSize", QueueSizeValue (GetMaxSize ()));
        for (uint8_t i=0; i<=Q; i++) 
        {
            AddInternalQueue (factory.Create<InternalQueue> ());
        }
    }
    if (GetNInternalQueues () != Q+1)
    {
        NS_LOG_ERROR ("TipsQueueDisc needs 3 internal queues");
        return false;
    }
    bool qs = false;
    for (size_t i = 0; i < Q+1; i++) {
    if (GetInternalQueue (i)-> GetMaxSize ().GetUnit () != QueueSizeUnit::PACKETS) {
        qs = true;
    }
    }
    if (qs)
    {
        NS_LOG_ERROR ("TipsQueueDisc needs 3 internal queues operating in packet mode");
        return false;
    }
    for (size_t i = 0; i < Q+1; i++)
    {
        if (GetInternalQueue (i)->GetMaxSize () < GetMaxSize ())
        {
            NS_LOG_ERROR ("The capacity of some internal queue(s) is less than the queue disc capacity");
            return false;
        }
    }
    return true;
}
void TipsQueueDisc::InitializeParams (void)
{
    NS_LOG_FUNCTION (this);
}
} // namespace ns3